-- Pulls back the Addon-Local Variables and store them locally.
local addonName = ...;
local addon = _G[addonName];

-- This is stored on the addon so its publically available to all other addons.
addon.Core.Tables = {
	ApplyDefaults = function(tableName, defaultValues)
		local Tables = addon.Core.Tables;
		local returnValue = Tables.DuplicateTables(tableName) or {};

		if not Tables.IsEmpty(defaultValues) then
			local pairs = pairs;

			for key, value in pairs(defaultValues) do
				if not returnValue[key] then
					returnValue[key] = value;
				elseif not Tables.IsEmpty(value) then
					returnValue[key] = Tables.ApplyDefaults(returnValue[key], value);
				end
			end
		end

		return returnValue;
	end,

	CaseInsensitive = function(...)
		local ipairs, strlower, pairs, setmetatable = ipairs, strlower, pairs, setmetatable;
		local Tables = addon.Core.Tables;
		local args = {...};

		-- Set the metatable to allow keys to be case insensitive.
		local metatable = {
			__mode = "k",

			__index = function(table, key)
				for k in pairs(table) do
					if strlower(k) == strlower(key) then
						return table[k];
					end
				end
			end,

			__metatable = "This table is protected!",
		};

		if not Tables.IsEmpty(args) then
			for _, table in ipairs(args) do
				-- Attach the metatable
				setmetatable(table, metatable);
			end
		else
			local returnValue = {};

			setmetatable(returnValue, metatable);

			return returnValue;
		end
	end,

	CompareTables = function(firstTable, secondTable)
		local Tables = addon.Core.Tables;

		if Tables.IsTable(firstTable) and Tables.IsTable(secondTable) then
			local pairs = pairs;

			for key, value in pairs(firstTable) do
				if Tables.IsTable(value) then
					if not Tables.CompareTables(value, secondTable[key]) then
						return false;
					end
				else
					if value ~= secondTable[key] then
						return false;
					end
				end
			end

			return true;
		end

		return firstTable == secondTable;
	end,

	CountTable = function(tableName)
		local pairs, type = pairs, type;
		local returnValue = 0;

		if type(tableName) == "table" then
			for _, _ in pairs(tableName) do
				returnValue = returnValue + 1;
			end
		end

		return returnValue;
	end,

	Dump = function(tableName)
		local format, pairs, print, strlen, strrep, tostring, type = format, pairs, print, strlen, strrep, tostring, type;
		local cachePrint = {};

		local function printTable(table, indent)
			if cachePrint[tostring(table)] then
				print(format("%s*%s", indent, tostring(table)));
			else
				cachePrint[tostring(table)] = true;

				if type(table) == "table" then
					for pos, val in pairs(table) do
						if type(val) == "table" then
							print(format("%s[%s] => %s {", indent, pos, tostring(table)));
							printTable(val, format("%s%s",indent, strrep(" ", strlen(pos) + 8)));
							print(format("%s%s}", indent, strrep(" ", strlen(pos) + 6)));
						else
							print(format("%s[%s] => %s", indent, pos, tostring(val)));
						end
					end
				else
					print(format("%s%s", indent, tostring(table)));
				end
			end
		end

		if type(tableName) == "table" then
			print(format("%s {", tostring(tableName)));
			printTable(tableName, "  " );
			print( "}" );
		else
			printTable(tableName, "  ");
		end
	end,

	DuplicateTables = function(tableName)
		local Tables = addon.Core.Tables;
		local returnValue = {};

		if not Tables.IsEmpty(tableName) then
			local pairs = pairs;

			for key, value in pairs(tableName) do
				if Tables.IsTable(value) then
					returnValue[key] = Tables.DuplicateTables(value);
				else
					returnValue[key] = value;
				end
			end
		end

		return returnValue;
	end,

	EquateTables = function(firstTable, secondTable)
		local Tables = addon.Core.Tables;
		local returnValue = firstTable == secondTable;

		if Tables.IsTable(firstTable) and Tables.IsTable(secondTable) then
			-- Compare the contents of the first table to the second table for missing or unequal values
			returnValue = Tables.CompareTables(firstTable, secondTable);

			-- Check to see if the contents were the same
			if returnValue then
				-- Compare the contents of the second table to the first table for missing values
				returnValue = Tables.CompareTables(secondTable, firstTable);
			end
		end

		return returnValue;
	end,

	-- Gets an enum table key from the value.
	GetKeyFromValue = function(enumTable, checkValue)
		local Tables = addon.Core.Tables;

		if not Tables.IsEmpty(enumTable) then
			local pairs = pairs;

			for key, value in pairs(enumTable) do
				if value == checkValue then
					return key;
				end
			end
		end

		return nil;
	end,

	GetValue = function(Object, property, default)
		local Tables = addon.Core.Tables;
		local returnValue = default;

		if not Tables.IsEmpty(Object) then
			local type = type;

			local propertyValue = Object[property];
			local valueType = type(propertyValue);

			if valueType == "function" then
				returnValue = propertyValue();
			elseif propertyValue ~= nil then
				returnValue = propertyValue;
			end
		end

		return returnValue;
	end,

	IsEmpty = function(tableName)
		local Tables = addon.Core.Tables;

		return Tables.CountTable(tableName) == 0;
	end,

	IsTable = function(tableName)
		local type = type;

		return (tableName and type(tableName) == "table") or false;
	end,

	MergeTables = function(firstTable, secondTable, noOverwrite)
		local Tables = addon.Core.Tables;
		local returnValue;

		if Tables.IsTable(firstTable) or Tables.IsTable(secondTable) then
			returnValue = Tables.DuplicateTables(firstTable);

			if not Tables.IsEmpty(secondTable) then
				local assert, format, pairs = assert, format, pairs;

				for key, value in pairs(secondTable) do
					assert(not noOverwrite or (noOverwrite and firstTable[key] == nil), format("Can't merge %s into first table as the key already exists!", key));

					if Tables.IsTable(value) then
						returnValue[key] = Tables.MergeTables(returnValue[key], value);
					else
						returnValue[key] = value;
					end
				end
			end
		end

		return returnValue;
	end,

	WipeTables = function(...)
		local ipairs, pairs, type = ipairs, pairs, type;

		local Tables = addon.Core.Tables;
		local args = {...};

		if not Tables.IsEmpty(args) then
			for _, table in ipairs(args) do
				if not Tables.IsEmpty(table) then
					for key, value in pairs(table) do
						if type(value) == "table" then
							Tables.WipeTables(value);
						else
							table[key] = nil;
						end
					end
				end
			end
		end
	end,
};