-- Pulls back the Addon-Local Variables and store them locally.
local addonName, addonTable = ...;
local addon = _G[addonName];

-- GUIDs used as keys for this table are summoned undead with limited duration.
local summonedCreatures = {};

-- SpellIDs associated with the tracked summons
-- Death Knight
local SPELLID_APOCALYPSE = 275699;
local SPELLID_ARMY_OF_THE_DEAD = 42650;
local SPELLID_GARGOYLE = 49206;
-- Shaman
local SPELLID_FERAL_SPIRIT = 262627;
local SPELLID_FIRE_ELEMENTAL = 198067;
local SPELLID_STORM_ELEMENTAL = 192249;

-- Creatures summoned
-- Death Knight
local CREATURE_APOCALYPSE = 999999;  -- Fake creatureID to seperate Army of the Dead Ghouls from Apocalypse Ghouls
local CREATURE_ARMY_OF_THE_DEAD = 24207;
local CREATURE_ARMY_OF_THE_DEAD_SPECIAL = 111101;  -- Special Army of the Dead spawn relating to Apocalypse Hidden Skin
local CREATURE_GARGOYLE = 27829;
-- Shaman
local CREATURE_FERAL_SPIRIT = 100820;
local CREATURE_FIRE_ELEMENTAL = 15438;
local CREATURE_STORM_ELEMENTAL = 77936;

-- Creature UnitIDs to check for summon events
local CLEU_SUMMON_CREATURES = {
	-- Death Knight
	[CREATURE_ARMY_OF_THE_DEAD] = true,
	[CREATURE_ARMY_OF_THE_DEAD_SPECIAL] = true,
	[CREATURE_GARGOYLE] = true,
	-- Shaman
	[CREATURE_FERAL_SPIRIT] = true,
	[CREATURE_FIRE_ELEMENTAL] = true,
	[CREATURE_STORM_ELEMENTAL] = true,
};

-- Summon duration for removing from the tracker
local SUMMONED_CREATURE_DURATIONS = {
	-- Death Knight
	[SPELLID_APOCALYPSE] = 15,
	[SPELLID_ARMY_OF_THE_DEAD] = 30,
	[SPELLID_GARGOYLE] = 40,
	-- Shaman
	[SPELLID_FERAL_SPIRIT] = 15,
	[SPELLID_FIRE_ELEMENTAL] = 30,
	[SPELLID_STORM_ELEMENTAL] = 30,
};

-- CreatureIDs that override the actual summoned creatureID
local SUMMONED_CREATURE_OVERRIDE_CREATUREID = {
	[SPELLID_APOCALYPSE] = CREATURE_APOCALYPSE,
	[SPELLID_ARMY_OF_THE_DEAD] = CREATURE_ARMY_OF_THE_DEAD,
};

local CLEU_UNIT_ADDED = {
	SPELL_SUMMON = true,
};

-- Table of CLEU events for when a unit is removed from combat.
local CLEU_UNIT_REMOVED = {
	UNIT_DESTROYED = true,
	UNIT_DIED = true,
	UNIT_DISSIPATES = true,
};

addon.Core.Objects.FinalizeSpellIDTable(SUMMONED_CREATURE_DURATIONS, SUMMONED_CREATURE_OVERRIDE_CREATUREID);

addonTable.Summons = {
	Active = function(ID)
		local pairs = pairs;
		local Tables = addon.Core.Tables;

		for _, data in pairs(summonedCreatures) do
			local creatureID = Tables.GetValue(data, "ID");
			local despawnTime = Tables.GetValue(data, "despawn");

			if (ID ~= nil and ID == creatureID) and (despawnTime ~= nil and addon.API.Blizzard.GetTime() < despawnTime) then
				return true;
			end
		end

		return false;
	end,

	Count = function(ID)
		local pairs = pairs;
		local Tables = addon.Core.Tables;
		local returnValue = 0;

		for _, data in pairs(summonedCreatures) do
			local creatureID = Tables.GetValue(data, "ID");
			local despawnTime = Tables.GetValue(data, "despawn");

			if (ID ~= nil and ID == creatureID) and (despawnTime ~= nil and addon.API.Blizzard.GetTime() < despawnTime) then
				returnValue = returnValue + 1;
			end
		end

		return returnValue;
	end,

	FirstDespawn = function(ID)
		local pairs = pairs;
		local Tables = addon.Core.Tables;
		local returnValue;

		for _, data in pairs(summonedCreatures) do
			local creatureID = Tables.GetValue(data, "ID");
			local despawnTime = Tables.GetValue(data, "despawn");

			if (ID ~= nil and ID == creatureID) and (despawnTime ~= nil and addon.API.Blizzard.GetTime() < despawnTime) then
				local remaining = despawnTime - addon.API.Blizzard.GetTime();

				if not returnValue or remaining < returnValue then
					returnValue = remaining;
				end
			end
		end

		return returnValue or 0;
	end,

	LastDespawn = function(ID)
		local pairs = pairs;
		local Tables = addon.Core.Tables;
		local returnValue;

		for _, data in pairs(summonedCreatures) do
			local creatureID = Tables.GetValue(data, "ID");
			local despawnTime = Tables.GetValue(data, "despawn");

			if (ID ~= nil and ID == creatureID) and (despawnTime ~= nil and addon.API.Blizzard.GetTime() < despawnTime) then
				local remaining = despawnTime - addon.API.Blizzard.GetTime();

				if not returnValue or remaining > returnValue then
					returnValue = remaining;
				end
			end
		end

		return returnValue or 0;
	end,
};

local function combatLog(...)
	local Misc = addon.Core.Misc;
	local Settings = addon.Core.Settings;
	local Tables = addon.Core.Tables;

	local currentTime = addon.API.Blizzard.GetTime();
	local _, _, combatEvent, _, sourceGUID, _, _, _, destGUID, _, _, _, spellID = ...;

	if sourceGUID == addon.API.Blizzard.UnitGUID("player") then
		local unitInfo = Misc.ParseGUID(destGUID);
		local creatureID = Tables.GetValue(unitInfo, "Creature")

		if CLEU_SUMMON_CREATURES[creatureID] then
			if CLEU_UNIT_REMOVED[combatEvent] then
				summonedCreatures[destGUID] = nil;
			elseif CLEU_UNIT_ADDED[combatEvent] then
				-- Check to see if the special Army of the Dead ghoul spawned for the Apocalypse Hidden Skin quest
				if Settings.GetRotationValue("UNHOLY_DEATHKNIGHT_HIDDENSKIN") and creatureID == CREATURE_ARMY_OF_THE_DEAD_SPECIAL then
					local ChatTypeInfo, RaidWarningFrame = ChatTypeInfo, RaidWarningFrame;

					-- Special ghoul spawned, so show a raid warning to bring it to the players attention!
					addon.API.Blizzard.RaidNotice_AddMessage(RaidWarningFrame, "Apocalyspe Hidden Skin Spawn Detected!", ChatTypeInfo["RAID_WARNING"]);
					addon.API.Blizzard.PlaySoundFile("Sound\\Creature\\LichKing\\IC_Lich King_Special01.ogg");
				end

				local spawnDuration = Tables.GetValue(SUMMONED_CREATURE_DURATIONS, spellID);
				local despawnTime = spawnDuration and currentTime + spawnDuration or nil;

				summonedCreatures[destGUID] = {
					ID = Tables.GetValue(SUMMONED_CREATURE_OVERRIDE_CREATUREID, spellID) or creatureID,
					spawn = currentTime,
					despawn = despawnTime,
				};
			end
		end
	end
end

-- Remove undead that have despawned without a combat log event.
local function timerEvent()
	local pairs = pairs;
	local Tables = addon.Core.Tables;
	local currentTime = addon.API.Blizzard.GetTime();

	for GUID, data in pairs(summonedCreatures) do
		local despawnTime = Tables.GetValue(data, "despawn");

		if despawnTime and currentTime > despawnTime then
			summonedCreatures[GUID] = nil;
		end
	end
end

local moduleEvents = {
	Disabled = function()
		local wipe = wipe;

		wipe(summonedCreatures);

		addon.System.Timers.TimerRemove(timerEvent);
	end,

	Enabled = function()
		-- Make sure the timer event is called immediately before starting the timer so it doesn't need to wait 1 second to get enemy data.
		timerEvent();

		addon.System.Timers.TimerAdd(timerEvent, 1);
	end,
};

local tableEvents = {
	COMBAT_LOG_EVENT_UNFILTERED = combatLog,
};

addon.System.Callbacks.RegisterAddonCallbacks(moduleEvents);
addon.System.Events.RegisterEvents(tableEvents, true);