-- Pulls back the Addon-Local Variables and store them locally.
local addonName, addonTable = ...;
local addon = _G[addonName];

-- Store local copy of Blizzard API wrapper and other addon global functions and variables
local Callbacks, Events = addon.System.Callbacks, addon.System.Events;
local error, format, setmetatable, sort, tinsert, type, wipe = error, format, setmetatable, sort, tinsert, type, wipe;

local rotationIndex = 0;
local actions = {};
local evaluateActions = {};

local function sortActions(a, b)
	if a.WaitTime == b.WaitTime then
		return a.RotationIndex < b.RotationIndex;
	end

	return a.WaitTime < b.WaitTime;
end

local function validInterrupt()
	local Settings = addon.Core.Settings;

	local Player = addon.Units.Player;
	local Pet = addon.Units.Pet;
	local groupType = (Player.IsSolo() and "Solo") or (Player.InParty() and "Party") or (Player.InRaid() and "Raid") or "";

	return Settings.GetCharacterValue("InterruptOptions", format("%s%s", "noLimits", groupType)) == 1
		or (Settings.GetCharacterValue("InterruptOptions", format("%s%s", "limitPlayer", groupType)) == 1 and Player.IsUnit("targettarget"))
		or (Settings.GetCharacterValue("InterruptOptions", format("%s%s", "limitPet", groupType)) == 1 and Pet.IsUnit("targettarget"))
		or (Settings.GetCharacterValue("InterruptOptions", format("%s%s", "limitFriendly", groupType)) == 1 and addon.Units.Target.IsFriendly("targettarget"));
end

local Prototype = {
	Actions = function()
		return actions;
	end,

	EvaluateInterruptAction = function(Object, useObject)
		local ActionBar = addon.Modules.ActionBar;
		local Objects = addon.Core.Objects;
		local Spell = addon.API.Spell;
		local Tables = addon.Core.Tables;

		local currentRotation = addon.currentRotation;
		local evaluateAction = Object.Evaluate(addon.Units.Target);

		rotationIndex = rotationIndex + 1;

		if Objects.IsSpell(Object) and Tables.GetValue(evaluateAction, "Use", false)
								   and Object.Charges() >= 1
								   and currentRotation.ProcessCondition(useObject) then
			local spellID = Object.SpellID();
			local objectInfo = Spell.GetSpellInfo(spellID);

			if not Tables.IsEmpty(objectInfo) then
				local iconInfo = {
					ID = spellID,
					Name = objectInfo.Name,
					Texture = objectInfo.IconID,
					Text = "",
					Type = "Spell",
					Keybind = ActionBar.Keybind("Spell", spellID),
					Wait = false,
					WaitStart = 0,
					WaitEnd = 0,
					WaitTime = 0;
					RotationIndex = rotationIndex,
				};

				tinsert(evaluateActions, iconInfo);
			end
		end
	end,
};

local ClassMT = {
	__index = Prototype,

	__newindex = function(table, key, value)
		error(format("Attempted to overwrite protected function %s in the Defensive Icon object!", key));
	end,

	__metatable = "This table is protected!",
};

addonTable.iconInterrupt = setmetatable({}, ClassMT);

local function timerEvent()
	local Debug = addonTable.Debug;
	local Tables = addon.Core.Tables;
	local Target = addon.Units.Target;

	local currentRotation = addon.currentRotation;
	local iconInterrupt = addonTable.iconInterrupt;

	if not Debug then
		if Target.IsValidTarget() and validInterrupt() then
			rotationIndex = 0;
			evaluateActions = {};

			if currentRotation then
				currentRotation.ProcessInterrupt(iconInterrupt);

				sort(evaluateActions, sortActions);
			end

			wipe(actions);
			actions = Tables.DuplicateTables(evaluateActions);

			Callbacks.Callback("InterruptIcon", actions);

			evaluateActions = nil;
		else
			wipe(actions);

			Callbacks.Callback("InterruptIcon");
		end
	end
end

local function checkCasting(...)
	local systemTimer = addon.System.Timers;
	local Target = addon.Units.Target;

	local _, unitID = ...;

	if not unitID or unitID == "target" then
		if Target.IsValidTarget() and Target.Casting.IsInterruptible() then
			if not systemTimer.TimerExists(timerEvent) then
				timerEvent();
				systemTimer.TimerAdd(timerEvent, addon.refreshInterval);
			end
		else
			if systemTimer.TimerExists(timerEvent) then
				systemTimer.TimerRemove(timerEvent);
				wipe(actions);

				Callbacks.Callback("InterruptIcon");
			end
		end
	end
end

local function startCasting(...)
	local systemTimer = addon.System.Timers;
	local Target = addon.Units.Target;

	local _, unitID = ...;

	if unitID == "target" then
		if Target.IsValidTarget() and Target.Casting.IsInterruptible() then
			if not systemTimer.TimerExists(timerEvent) then
				timerEvent();
				systemTimer.TimerAdd(timerEvent, addon.refreshInterval);
			end
		else
			if systemTimer.TimerExists(timerEvent) then
				systemTimer.TimerRemove(timerEvent);
				wipe(actions);

				Callbacks.Callback("InterruptIcon");
			end
		end
	end
end

local function stopCasting(...)
	local systemTimer = addon.System.Timers;
	local Target = addon.Units.Target;

	local _, unitID = ...;

	if unitID == "target" then
		if not Target.IsValidTarget() or not Target.Casting.IsInterruptible() then
			if systemTimer.TimerExists(timerEvent) then
				systemTimer.TimerRemove(timerEvent);

				wipe(actions);

				Callbacks.Callback("InterruptIcon");
			end
		end
	end
end

local moduleEvents = {
	Disabled = function()
		local systemTimer = addon.System.Timers;

		systemTimer.TimerRemove(timerEvent);
		wipe(actions);

		Callbacks.Callback("InterruptIcon");
	end,

	Enabled = function()
		checkCasting();
	end,
};

local tableEvents = {
	PLAYER_TARGET_CHANGED = checkCasting,
	UNIT_SPELLCAST_CHANNEL_START = startCasting,
	UNIT_SPELLCAST_CHANNEL_STOP = stopCasting,
	UNIT_SPELLCAST_CHANNEL_UPDATE = startCasting,
	UNIT_SPELLCAST_FAILED = stopCasting,
	UNIT_SPELLCAST_FAILED_QUIET = stopCasting,
	UNIT_SPELLCAST_INTERRUPTED = stopCasting,
	UNIT_SPELLCAST_START = startCasting,
	UNIT_SPELLCAST_STOP = stopCasting,
	UNIT_SPELLCAST_SUCCEEDED = checkCasting,
};

Callbacks.RegisterAddonCallbacks(moduleEvents);
Events.RegisterEvents(tableEvents);