-- Pulls back the Addon-Local Variables and store them locally.
local _, addonTable = ...;
local addon = _G["LunaEclipse_Rotation"];

-- Plugin information variables.
local pluginName = "Weak Auras";

-- Check whether the plugin is supposed to be active and that the required Weak Auras is loaded.
if addon and addon.API.Blizzard.IsAddOnLoaded("WeakAuras") and addon.System.Plugins.IsActive(pluginName) then
	-- Store local copies of Blizzard API and other addon global functions and variables
	local ipairs = ipairs;

	-- Local variables for plugin operation.
	local registeredStyles = {};

	-- Tables to hold private functions.
	local Auras, Styles = {}, {};

	--------------------- Private Module Functions ---------------
	Auras = {
		Add = function(aura)
			local Tables = addon.Core.Tables;

			if not Tables.IsEmpty(aura) then
				local parentName = Tables.GetValue(aura, "parent", "");
				local parentData = parentName ~= "" and WeakAuras.GetData(parentName);

				WeakAuras.Add(aura);
				WeakAuras.SyncParentChildRelationships(true);

				if WeakAuras.OptionsFrame then
					WeakAuras.NewDisplayButton(aura);

					if parentData then
						local button = WeakAuras.GetDisplayButton(parentName);

						if button then
							button.callbacks.UpdateExpandButton();
						end

						WeakAuras.UpdateDisplayButton(parentData);
						WeakAuras.ReloadGroupRegionOptions(parentData);
					end
				end
			end
		end,

		Build = function(aura, styleName)
			local Tables = addon.Core.Tables;
			local auraName = Tables.GetValue(aura.ID, "id");

			if auraName then
				local customAuras = Tables.GetValue(addonTable.weakAuraSettings.CustomAuras, styleName, {});
				local customAura = Tables.GetValue(customAuras, auraName);
				local originalAura = WeakAuras.GetData(auraName);

				if not Tables.IsEmpty(originalAura) then
					return Auras.Modify(originalAura, aura);
				elseif not Tables.IsEmpty(customAura) then
					return Auras.Modify(customAura, aura);
				else
					return Auras.Default(aura);
				end
			end

			return nil;
		end,

		Default = function(aura)
			local Tables = addon.Core.Tables;
			local returnValue;

			if not Tables.IsEmpty(aura) then
				local templates = addonTable.WeakAuras.Templates;

				if Auras.IsIcon(aura) == 1 then
					returnValue = templates.templateIcon;
				else
					returnValue = templates.templateText;
				end

				returnValue = Tables.MergeTables(returnValue, aura.ID);
				returnValue = Tables.MergeTables(returnValue, aura.Position);
				returnValue = Tables.MergeTables(returnValue, aura.CustomLua);
			end

			return returnValue;
		end,

		Delete = function(aura)
			local Tables = addon.Core.Tables;

			if not Tables.IsEmpty(aura) then
				if WeakAuras.OptionsFrame then
					local parentName = Tables.GetValue(aura, "parent");

					WeakAuras.DeleteOption(aura);

					if parentName then
						local parentData = WeakAuras.GetData(parentName);
						local parentButton = WeakAuras.GetDisplayButton(parentName);

						if parentData then
							WeakAuras.UpdateGroupOrders(parentData);
						end

						if parentButton then
							parentButton.callbacks.UpdateExpandButton();
						end
					end
				else
					WeakAuras.Delete(aura);
					WeakAuras.SyncParentChildRelationships(true);
				end
			end
		end,

		IsIcon = function(aura)
			local Tables = addon.Core.Tables;

			return Tables.GetValue(aura.ID, "regionType", "") == "icon" and 1 or 0;
		end,

		Modify = function(aura, updateAura)
			local Tables = addon.Core.Tables;
			local returnValue;

			if not Tables.IsEmpty(aura) then
				returnValue = Tables.DuplicateTables(aura);

				if not Tables.IsEmpty(updateAura) then
					returnValue = Tables.MergeTables(aura, updateAura.CustomLua);
				end
			end

			return returnValue;
		end,
	};

	Styles = {
		Apply = function(styleName)
			if styleName then
				local Core = addonTable.WeakAuras.Core;
				local Tables = addon.Core.Tables;

				local addonSettings = addonTable.addonSettings;
				local styleData = Tables.GetValue(registeredStyles, styleName);

				if not Tables.IsEmpty(styleData) then
					local parentData = Tables.GetValue(styleData, "Parent");
					local auraImports = Tables.GetValue(styleData, "Imports");

					if not Tables.IsEmpty(parentData) and not Tables.IsEmpty(auraImports) then
						Core.AddAura(parentData, styleName);

						for _, weakAura in ipairs(auraImports) do
							Core.AddAura(weakAura, styleName);
						end

						addonSettings.versionWeakAura = addonTable.WeakAuras.Version;
					end
				end
			end
		end,

		Register = function(styleName, weakAurasOptions, parentAura, auraImports)
			if styleName then
				local Tables = addon.Core.Tables;
				local styleData = Tables.GetValue(registeredStyles, styleName);

				if Tables.IsEmpty(styleData) then
					registeredStyles[styleName] = {
						Options = weakAurasOptions,
						Parent = parentAura,
						Imports = auraImports,
					};
				end
			end
		end,

		Remove = function(styleName)
			if styleName then
				local Core = addonTable.WeakAuras.Core;
				local Tables = addon.Core.Tables;

				local currentStyleName = Tables.GetValue(addonTable.addonSettings, "styleSelection");
				local styleData = Tables.GetValue(registeredStyles, styleName);

				if not Tables.IsEmpty(styleData) then
					local parentData = Tables.GetValue(styleData, "Parent");
					local auraImports = Tables.GetValue(styleData, "Imports");

					if not Tables.IsEmpty(parentData) and not Tables.IsEmpty(auraImports) and styleName ~= currentStyleName then
						local auraID = Tables.GetValue(parentData, "id");
						local region = WeakAuras.regions[auraID];

						if not Tables.IsEmpty(region.ControlChildren) then
							region:Pause();
						end

						for _, weakAura in ipairs(auraImports) do
							Core.RemoveAura(weakAura, styleName);
						end

						Core.RemoveAura(parentData, styleName);
					end
				end
			end
		end,
	};

	------------- Private Internal Namespace Functions -----------
	addonTable.WeakAuras.Core = {
		AddAura = function(aura, styleName)
			if not WeakAuras.IsImporting() and styleName then
				local Tables = addon.Core.Tables;
				local auraName = Tables.GetValue(aura, "id");

				if auraName then
					-- This is the parent group, so just add it if its not already existing.
					local auraData = WeakAuras.GetData(auraName);

					if Tables.IsEmpty(auraData) then
						Auras.Add(aura);
					end
				else
					auraName = Tables.GetValue(aura.ID, "id");

					local profile = Tables.GetValue(addonTable.weakAuraSettings.Styles, styleName, {});
					local auraEnabled = Tables.GetValue(profile, auraName, Auras.IsIcon(aura)) == 1;

					local auraData = WeakAuras.GetData(auraName);
					local newAura = Auras.Build(aura, styleName);

					if auraEnabled and (Tables.IsEmpty(auraData) or Tables.GetValue(addonTable.addonSettings, "versionWeakAura") < addonTable.WeakAuras.Version) then
						Auras.Add(newAura);
					elseif not auraEnabled and not Tables.IsEmpty(auraData) then
						Auras.Delete(auraData);
					end
				end
			end
		end,

		RemoveAura = function(aura, styleName)
			if not WeakAuras.IsImporting() and styleName then
				local Tables = addon.Core.Tables;
				local auraName = Tables.GetValue(aura, "id") or Tables.GetValue(aura.ID, "id");

				if auraName then
					local auraData = WeakAuras.GetData(auraName);

					if not Tables.IsEmpty(auraData) then
						local profile = Tables.GetValue(addonTable.weakAuraSettings.CustomAuras, styleName, {});

						if not Tables.EquateTables(auraData, Auras.Default(aura)) then
							-- Aura was modified so be sure to back it up
							profile[auraName] = auraData;

							-- Make sure we save any aura customization to saved variables
							addonTable.weakAuraSettings.CustomAuras[styleName] = profile;
						end

						-- Delete the child aura from the group.
						Auras.Delete(auraData);
					end
				end
			end
		end,

		SetupConfiguration = function(styleName)
			if styleName then
				local Options = addon.Options.Plugins;
				local Tables = addon.Core.Tables;

				local WeakAurasOptions = addon.Options.pluginOptions.args[Options.CovertPluginName(pluginName)];
				local styleData = Tables.GetValue(registeredStyles, styleName);

				if not Tables.IsEmpty(styleData) then
					local optionsTable = Tables.GetValue(styleData, "Options");

					WeakAurasOptions.args.RotationIcons.args = Tables.DuplicateTables(optionsTable.RotationIcons.args);
					WeakAurasOptions.args.InfoTexts.args = Tables.DuplicateTables(optionsTable.InfoTexts.args);

					addon.System.Callbacks.Callback("optionsChanged", "LERH_PluginSettings");
				end
			end
		end,
	};

	---------------------- Public Functions ----------------------
	addon.Features.WeakAuras = {
		ApplyStyle = function(styleName)
			addon.System.Events.CombatEventDelay(Styles.Apply, styleName);
		end,

		RegisterStyle = function(styleName, weakAurasOptions, parentAura, auraImports)
			Styles.Register(styleName, weakAurasOptions, parentAura, auraImports);
		end,

		RemoveStyle = function(styleName)
			addon.System.Events.CombatEventDelay(Styles.Remove, styleName);
		end,
	};

	----------------------- Plugin Code Operation ----------------
	local moduleEvents = {
		playerLogin = function()
			local Tables = addon.Core.Tables;

			local addonSettings = addonTable.addonSettings;
			local styleName = Tables.GetValue(addonSettings, "styleSelection", "Ovale");

			addonTable.WeakAuras.Core.SetupConfiguration(styleName);
			addon.Features.WeakAuras.ApplyStyle(styleName);
		end,

		CooldownIcon = function(ActionList)
			WeakAuras.ScanEvents("LERH_CD_UPDATE", ActionList);
		end,

		DefensiveIcon = function(ActionList)
			WeakAuras.ScanEvents("LERH_DEFENSIVE_UPDATE", ActionList);
		end,

		InterruptIcon = function(ActionList)
			WeakAuras.ScanEvents("LERH_INTERRUPT_UPDATE", ActionList);
		end,

		OffGCDIcon = function(ActionList)
			WeakAuras.ScanEvents("LERH_OFFGCD_UPDATE", ActionList);
		end,

		MultiTargetIcon = function(ActionList)
			WeakAuras.ScanEvents("LERH_AOE_UPDATE", ActionList);
		end,

		SingleTargetIcon = function(ActionList)
			WeakAuras.ScanEvents("LERH_ST_UPDATE", ActionList);
		end,
	};

	-- These are only registered if the plugin is active as they control how what is called
	-- from LunaEclipse Rotation Helper
	addon.System.Callbacks.RegisterAddonCallbacks(moduleEvents);
end