-- Pulls back the Addon-Local Variables and store them locally.
local _, addonTable = ...;
local addon = _G["LunaEclipse_Rotation"];

-- Plugin information variables.
local pluginName = "Weak Auras";
local pluginDescription = "This plugin automatically creates and updates the Weak Auras needed for the Rotation Helper addon.\n\nThis plugin is also responsible for sending the custom event to Weak Auras so the auras will update the suggested ability.";

-- Check for WeakAuras and LunaEclipse Rotation Helper being loaded first so it doesn't register the plugin if its
-- not needed.
if addon and addon.API.Blizzard.IsAddOnLoaded("WeakAuras") then
    -- Store local copies of Blizzard API and other addon global functions and variables
    local pairs = pairs;

    local optionsTable = {
        type = "group",
        name = pluginName,
        childGroups = "tree",
        args = {
            description = {
                type = "description",
                name = "These settings affect which Weak Auras are displayed for the user interface.",
                width = "full",
                order = 10,
            },
            styleSelection = {
                type = "select",
                style = "dropdown",
                name = "WeakAuras Layout",
                values = function()
                    local returnValue = {};

                    for key in pairs(addonTable.WeakAuras.Styles) do
                        returnValue[key] = key;
                    end

                    return returnValue;
                end,
                get = function(info)
                    local returnValue = "Ovale";

                    if info then
                        local Tables = addon.Core.Tables;
                        local profile = addonTable.addonSettings;
                        local keyName = info[#info];

                        returnValue = Tables.GetValue(profile, keyName);

                        if returnValue == nil then
                            returnValue = "Ovale";
                            profile[keyName] = returnValue;

                            addonTable.addonSettings = profile;
                        end
                    end

                    return returnValue;
                end,
                set = function(info, value)
                    if info then
                        local Tables = addon.Core.Tables;
                        local profile = addonTable.addonSettings;
                        local keyName = info[#info];
                        local previousStyle = Tables.GetValue(profile, keyName);

                        profile[keyName] = value;

                        -- Put the settings into the database
                        addonTable.addonSettings = profile;

                        if value ~= nil then
                            addonTable.WeakAuras.Core.SetupConfiguration(value);

                            addon.Features.WeakAuras.RemoveStyle(previousStyle);
                            addon.Features.WeakAuras.ApplyStyle(value);
                        end
                    end
                end,
                width = "double",
                order = 20,
            },
            RotationIcons = {
                type = "group",
                name = addon.Options.Core.CreaterHeader("Rotation Icons"),
                get = function(info)
                    local returnValue = 1;

                    if info then
                        local Tables = addon.Core.Tables;
                        local addonSettings = addonTable.addonSettings;
                        local weakAuraSettings = addonTable.weakAuraSettings;

                        local styleName = Tables.GetValue(addonSettings, "styleSelection", "Ovale");
                        local profile = Tables.GetValue(weakAuraSettings.Styles, styleName, {});
                        local keyName = info[#info];

                        returnValue = Tables.GetValue(profile, keyName);

                        if returnValue == nil then
                            returnValue = 1;
                            profile[keyName] = returnValue;

                           addonTable.weakAuraSettings.Styles[styleName] = profile;
                        end

                        if info.option.type == "toggle" then
                            returnValue = addon.Options.Core.ConvertToUI(returnValue, info.option.tristate);
                        end
                    end

                    return returnValue;
                end,
                set = function(info, value)
                    if info then
                        local Tables = addon.Core.Tables;
                        local addonSettings = addonTable.addonSettings;
                        local weakAuraSettings = addonTable.weakAuraSettings;

                        local styleName = Tables.GetValue(addonSettings, "styleSelection", "Ovale");
                        local profile = Tables.GetValue(weakAuraSettings.Styles, styleName, {});
                        local keyName = info[#info];

                        if info.option.type == "toggle" then
                            value = addon.Options.Core.ConvertToValue(value, info.option.tristate);
                        end

                        profile[keyName] = value;

                        -- Put the settings into the database
                        addonTable.weakAuraSettings.Styles[styleName] = profile;

                        -- Apply the aura change.
                        addonTable.WeakAuras.Core.AddAura(addonTable.WeakAuras.Styles[styleName].Data.Icons[keyName], styleName);
                        WeakAuras.SyncParentChildRelationships(true);
                    end
                end,
                order = 30,
                inline = true,
                args = {},
            },
            InfoTexts = {
                type = "group",
                name = addon.Options.Core.CreaterHeader("Text Information Displays"),
                get = function(info)
                    local returnValue = 0;

                    if info then
                        local Tables = addon.Core.Tables;
                        local addonSettings = addonTable.addonSettings;
                        local weakAuraSettings = addonTable.weakAuraSettings;

                        local styleName = Tables.GetValue(addonSettings, "styleSelection", "Ovale");
                        local profile = Tables.GetValue(weakAuraSettings.Styles, styleName, {});
                        local keyName = info[#info];

                        returnValue = Tables.GetValue(profile, keyName);

                        if returnValue == nil then
                            returnValue = 0;
                            profile[keyName] = returnValue;

                            addonTable.weakAuraSettings.Styles[styleName] = profile;
                        end

                        if info.option.type == "toggle" then
                            returnValue = addon.Options.Core.ConvertToUI(returnValue, info.option.tristate);
                        end
                    end

                    return returnValue;
                end,
                set = function(info, value)
                    if info then
                        local Tables = addon.Core.Tables;
                        local addonSettings = addonTable.addonSettings;
                        local weakAuraSettings = addonTable.weakAuraSettings;

                        local styleName = Tables.GetValue(addonSettings, "styleSelection", "Ovale");
                        local profile = Tables.GetValue(weakAuraSettings.Styles, styleName, {});
                        local keyName = info[#info];

                        if info.option.type == "toggle" then
                            value = addon.Options.Core.ConvertToValue(value, info.option.tristate);
                        end

                        profile[keyName] = value;

                        -- Put the settings into the database
                        addonTable.weakAuraSettings.Styles[styleName] = profile;

                        -- Apply the aura change.
                        addonTable.WeakAuras.Core.AddAura(addonTable.WeakAuras.Styles[styleName].Data.Texts[keyName], styleName);
                        WeakAuras.SyncParentChildRelationships(true);
                    end
                end,
                order = 40,
                inline = true,
                args = {},
            },
        },
    };

    -- The plugin still has to be registered even if its not loaded into memory, otherwise there is no checkbox to re-enable it.
    addon.System.Plugins.RegisterPlugin(pluginName, pluginDescription, optionsTable);
end